/*
 * yasapi.h
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#ifndef __YASAPI_H__ // {
#define __YASAPI_H__
#include <ya.h>
#include <yasapi_guid.h>
#ifdef __cpluplus
extern "C" {
#endif

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_TIME_SEGMENT
#define YASAPI_GAPLESS
#define YASAPI_SORROUND
#define YASAPI_RING_REALLOC
#define YASAPI_RING_DWORD
#define YASAPI_RING_BUF_SIZE
#define YASAPI_NOTIFY
//#define YASAPI_CO_INITIALIZE
//#define YASAPI_CHANNEL_LAYOUT

#if defined (YASAPI_GAPLESS) // {
//#define YASAPI_CHECK_UNDERFLOW
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if ! defined (YASAPI_VER) // {
  #define YASAPI_VER            0.0.0
#endif // }

#define YASAPI_VERSION          YA_STR(YASAPI_VER)
#define PI_VER2                 "v" YASAPI_VERSION

#if defined (OUT_YASAPI_SFX) // {
  #define YASAPI_PREFIX         YASAPI_CONCAT(out_yasapi,OUT_YASAPI_SFX)
#else // } {
  #define YASAPI_PREFIX         out_yasapi
#endif // }

#define YASAPI_ID_BASE          65536
#define YASAPI_ID_OFFS          858

#if defined (YA_DEBUG) && defined (YA_SSE2) // {
  #define PI_VER                PI_VER2 " (Debug/SSE2)"
  #define YASAPI_ID             YASAPI_PREFIX ## -debug-sse2
  #define YASAPI_WA_ID          (YASAPI_ID_BASE+YASAPI_ID_OFFS+3)
#elif defined (YA_DEBUG) // {
  #define PI_VER                PI_VER2 " (Debug)"
  #define YASAPI_ID             YASAPI_PREFIX ## -debug
  #define YASAPI_WA_ID          (YASAPI_ID_BASE+YASAPI_ID_OFFS+0)
#elif defined (YA_SSE2) // } {
  #define PI_VER                PI_VER2 " (SSE2)"
  #define YASAPI_ID             YASAPI_PREFIX ## -sse2
  #define YASAPI_WA_ID          (YASAPI_ID_BASE+YASAPI_ID_OFFS+1)
#else // } {
  #define PI_VER                PI_VER2
  #define YASAPI_ID             YASAPI_PREFIX
  #define YASAPI_WA_ID          (YASAPI_ID_BASE+YASAPI_ID_OFFS+2)
#endif // }

#define YASAPI_NAME             YA_STR(YASAPI_ID)
#define YASAPI_MODULE           YASAPI_NAME ".dll"
#define YASAPI_PROPERTY_FILE_NAME \
                                L"Plugins\\" YA_WSTR(YASAPI_ID) L".ini"
#define PI_LABEL                "Yet Another (WA)SAPI Output " PI_VER

#define YASAPI_SHAREMODE_LABEL(eShareMode) \
    (AUDCLNT_SHAREMODE_SHARED==(eShareMode)?"SHARED":"EXCLUSIVE")

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_FRAMES_MS(nFrames,pwfx) \
  MulDiv(1000,nFrames,(pwfx)->nSamplesPerSec)
#define YASAPI_FRAMES_TIMER(nFrames,pwfx) \
  (10000.0*1000*(nFrames)/(pwfx)->nSamplesPerSec)

#define YASAPI_MIN_BALANCE_COUNT  (8ull)

#define YASAPI_MIN_RING_SIZE      (1.0)
#define YASAPI_MAX_RING_SIZE      (10.0)

#define YASAPI_MIN_SHARE_SIZE     (1.0)
#define YASAPI_MAX_SHARE_SIZE     (20.0)

#if defined (YASAPI_UNDERFLOW) // {
#define YASAPI_MIN_UNDERFLOW      (0.0)
#define YASAPI_MAX_UNDERFLOW      (1.0)
#endif // }

///////////////////////////////////////////////////////////////////////////////
DEFINE_GUID(IID_IUnknown, 0x00000000, 0x0000, 0x0000, 0xc0,0x00,
    0x00,0x00,0x00,0x00,0x00,0x46);

DEFINE_GUID(IID_IAudioClient, 0x1cb9ad4c, 0xdbfa, 0x4c32, 0xb1,0x78,
    0xc2,0xf5,0x68,0xa7,0x03,0xb2);
DEFINE_GUID(IID_IAudioClock, 0xcd63314f, 0x3fba, 0x4a1b, 0x81,0x2c,
    0xef,0x96,0x35,0x87,0x28,0xe7);
DEFINE_GUID(IID_IAudioRenderClient, 0xf294acfc, 0x3146, 0x4483, 0xa7,0xbf,
    0xad,0xdc,0xa7,0xc2,0x60,0xe2);

DEFINE_GUID(CLSID_MMDeviceEnumerator, 0xbcde0395, 0xe52f, 0x467c, 0x8e,0x3d,
    0xc4,0x57,0x92,0x91,0x69,0x2e);
DEFINE_GUID(IID_IMMDeviceEnumerator, 0xa95664d2, 0x9614, 0x4f35, 0xa7,0x46,
    0xde,0x8d,0xb6,0x36,0x17,0xe6);
DEFINE_GUID(IID_IMMDevice, 0xd666063f, 0x1587, 0x4e43, 0x81,0xf1,
    0xb9,0x48,0xe8,0x07,0x36,0x3f);

DEFINE_GUID(IID_IMMNotificationClient, 0x7991eec9, 0x7e89, 0x4d85, 0x83,0x90,
    0x6c,0x70,0x3c,0xec,0x60,0xc0);
#if 0 // {
DEFINE_GUID(CLSID_MMPlayerNotificationClient, 0x1e45b183,0x7929,0x40e1,
    0xB1,0x9B,0x93,0x83,0x2A,0x44,0x49,0x24);
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_CHANNEL_LAYOUT) // {
typedef enum _ChannelLayout ChannelLayout;
#endif // }
typedef enum _PlayerState PlayerState;

typedef enum _TimeTag TimeTag;
typedef struct _OptionsCommon OptionsCommon;
typedef struct _OptionsDevice OptionsDevice;
typedef struct _Options Options;

typedef struct _Connection Connection;
typedef union _TimeSegment TimeSegment;
typedef struct _Time Time;
typedef struct _PlayerDevice PlayerDevice;
typedef struct _Player Player;
typedef struct _Strategy Strategy;

///////////////////////////////////////////////////////////////////////////////
#define VALUE_INT(pData,offs) \
  (*(int *)(((char *)(pData))+(offs)))
#define VALUE_DOUBLE(pData,offs) \
  (*(double *)(((char *)(pData))+(offs)))
#define VALUE_STRING(pData,offs) \
  ((wchar_t *)(((char *)(pData))+(offs)))

#define PROPERTY_INT(pProperty,pData) \
  VALUE_INT(pData,(pProperty)->offs)
#define PROPERTY_DOUBLE(pProperty,pData) \
  VALUE_DOUBLE(pData,(pProperty)->offs)
#define PROPERTY_STRING(pProperty,pData) \
  VALUE_STRING(pData,(pProperty)->offs)

#define YASAPI_PROPERTY_SIZE    128
#define YASAPI_GROUP_COMMON     L"plugin"
#define YASAPI_PROPERTY_DEBUG   L"debug"
#define YASAPI_PROPERTY_VIS     L"visualization"
#define YASAPI_PROPERTY_PAGE    L"page"
#define YASAPI_PROPERTY_POSX    L"posx"
#define YASAPI_PROPERTY_POSY    L"posy"

#if defined (YASAPI_CHANNEL_LAYOUT) // {
///////////////////////////////////////////////////////////////////////////////
enum _ChannelLayout {
  CHANNEL_LAYOUT_NULL,
  CHANNEL_LAYOUT_MONO,
  CHANNEL_LAYOUT_STEREO,
  CHANNEL_LAYOUT_3POINT1,             // CHANNEL_LAYOUT_QUAD
  CHANNEL_LAYOUT_3POINT1_SURROUND,    // CHANNEL_LAYOUT_SURROUND
  CHANNEL_LAYOUT_5POINT1,
  CHANNEL_LAYOUT_5POINT1_SORROUND,
  CHANNEL_LAYOUT_7POINT1,
  CHANNEL_LAYOUT_7POINT1_SURROUND,
  CHANNEL_LAYOUT_9POINT1,
  CHANNEL_LAYOUT_9POINT1_SURROUND
};
#endif // }

///////////////////////////////////////////////////////////////////////////////
typedef void (*RingCopy)(PVOID pClient, PVOID Destination,
    const VOID *Source, SIZE_T Length);

enum {
  RING_COPY=1,
  RING_COMMIT=2
};

struct _Ring {
  RingCopy Copy;
  PVOID pClient;
  LPSTR pRep;
  LPSTR pMax;
  LPSTR pHead;
  LPSTR pTail;
  SIZE_T dwWritten;
  SIZE_T dwAvailable;
};

DWORD RingCreate(Ring *pRing, SIZE_T dwSize);
void RingDestroy(Ring *pRing);

void RingReset(Ring *pRing);
#if defined (YASAPI_RING_REALLOC) // {
int RingRealloc(Ring *pRing, SIZE_T uSize);
int RingReallocAvailable(Ring *pRing, SIZE_T dwAvailable, SIZE_T n);
#endif // }
#if defined (YASAPI_RING_DWORD) // {
DWORD RingWrite(Ring *pRing, LPCSTR pData, SIZE_T dwSize, SIZE_T n);
DWORD RingReadEx(Ring *pRing, LPSTR pData, SIZE_T dwSize, UINT uFlags);
DWORD RingRead(Ring *pRing, LPSTR pData, SIZE_T dwSize);
#else // } {
SIZE_T RingWrite(Ring *pRing, LPCSTR pData, SIZE_T dwSize, SIZE_T n);
SIZE_T RingReadEx(Ring *pRing, LPSTR pData, SIZE_T dwSize, UINT uFlags);
SIZE_T RingRead(Ring *pRing, LPSTR pData, SIZE_T dwSize);
#endif // }

void RingCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length);

///////////////////////////////////////////////////////////////////////////////
enum _OptionsSize {
  YASAPI_ID_SIZE=128
};

enum _TimeTag {
  TIME_POSITION,
  TIME_TIME
};

struct _OptionsCommon {
  // Persistent fields have to be of type "int", "double" or "wchar_t[]".
  wchar_t szId[YASAPI_ID_SIZE];
  int bMono2Stereo;
  int bVolume;
#if defined (YASAPI_SORROUND) // {
  int bSorround;
#endif // }
#if defined (YASAPI_GAPLESS) // {
  int bGapless;
  int bDisconnect;
  TimeTag eTimeTag;
#if defined (YASAPI_CHECK_UNDERFLOW) // {
  int nCheckUnderflow;
#endif // }
#else // } {
  int nWriteBlock;
#endif // }
  int bFormatSupported;
  int bAudioClock;
  int bVisualization;
  int nPage;
#if defined (YASAPI_ABOUT) // {
  int nConfigX;
  int nConfigY;
  int nAboutX;
  int nAboutY;
#else // } {
  int nPosX;
  int nPosY;
#endif // }
};

int OptionsCommonLoad(OptionsCommon *pOptions, const wchar_t *pfx,
    const wchar_t *path);
void OptionsCommonSave(OptionsCommon *pOptions, const wchar_t *path);
const OptionsCommon *OptionsCommonDefault(void);

///////////////////////////////////////////////////////////////////////////////
enum _OptionsSharemode {
  YASAPI_SHAREMODE_SHARE,
  YASAPI_SHAREMODE_EXCLUSIVE,
  YASAPI_SHAREMODE_AUTOMATIC
};

enum _OptionsDevicePeriod {
  YASAPI_DEVICE_PERIOD_DEFAULT,
  YASAPI_DEVICE_PERIOD_MINIMUM,
};

struct _OptionsDevice {
  // Persistent fields have to be of type "int", "double" or "wchar_t[]".
  int eShareMode;
  int eDevicePeriod;
  int bAutoConvertPCM;
  int bSRCDefaultQuality;
  int bPull;
  double qShareSize;

  struct {
    double qSize;
    double qFill;
  } ring;
};

int OptionsDeviceLoad(OptionsDevice *pOptions, const wchar_t *pfx,
    const wchar_t *pstrId, const wchar_t *path);
void OptionsDeviceSave(OptionsDevice *pOptions, const wchar_t *pstrId,
    const wchar_t *path);
const OptionsDevice *OptionsDeviceDefault(void);

///////////////////////////////////////////////////////////////////////////////
struct _Options {
  const wchar_t *path;
  OptionsCommon common;
  OptionsDevice device;
};

///////////////////////////////////////////////////////////////////////////////
struct _Connection {
  int bInvalid;
  UINT32 uFrames;
  UINT32 uFramesMin;
  LONGLONG time;
  IAudioClient *pClient;
  IAudioRenderClient *pRender;
  IAudioClock *pClock;
};

int ConnectionGetPosition(Connection *pConnect, UINT64 *pu64Position);
int ConnectionGetFrequency(Connection *pConnect, UINT64 *pu64Frequency);

///////////////////////////////////////////////////////////////////////////////
union _TimeSegment {
  UINT64 u64Position;
  double xSeconds;
};

void TimeSegmentReset(TimeSegment *pSegment, UINT64 u64Frequency);
void TimeSegmentAdd(TimeSegment *pOffset, TimeSegment *pCurrent,
    UINT64 u64Frequency);
#if 0 // {
void TimeSegmentSub(TimeSegment *pOffset, TimeSegment *pCurrent,
    UINT64 u64Frequency);
#endif // }
int TimeSegmentGet(TimeSegment *pSegment, UINT64 u64Frequency,
    Connection *pConnect);

struct _Time {
  // time is pause + current - gapless
  UINT64 u64Frequency;
  TimeSegment gapless;
  TimeSegment pause;
  TimeSegment current;
};

int TimeReset(Time *pTime, TimeTag eTimeTag, Connection *pConnect);
int TimePause(Time *pTime, Connection *pConnect);
void TimeFlush(Time *pTime);
int TimeGapless(Time *pTime, Connection *pConnect);
int TimeMigrate(Time *pTime, Connection *pConnect);
int TimeGetMS(Time *pTime, Connection *pConnect, double *ms);

///////////////////////////////////////////////////////////////////////////////
struct _PlayerDevice {
  IMMDevice *pDevice;
  wchar_t szId[YASAPI_ID_SIZE];
};

int PlayerDeviceCreate(PlayerDevice *pPlayerDevice, PWSTR pstrId,
    IMMDeviceEnumerator *pEnumerator);
int PlayerDeviceDestroy(PlayerDevice *pPlayerDevice);
int PlayerDeviceGet(PlayerDevice *pPlayerDevice,
    IMMDeviceEnumerator *pEnumerator);

int64_t PlayerDeviceCreateV(Player *pPlayer, Request *pRequst);
int64_t PlayerDeviceDestroyV(Player *pPlayer, Request *pRequst);

///////////////////////////////////////////////////////////////////////////////
enum _PlayerState {
  PLAYER_STATE_NULL,
  PLAYER_STATE_BASE,
  PLAYER_STATE_RUN,
  PLAYER_STATE_PAUSE,
  PLAYER_STATE_UNDERFLOW,
  PLAYER_STATE_READY,
  PLAYER_STATE_PLAY,
  // 1. client indicates draining.
  PLAYER_STATE_DRAIN_1,
  // 2. ring buffer is empty (drained).
  PLAYER_STATE_DRAIN_2,
  // 3. player awaits read request in order to switch to eot.
  PLAYER_STATE_DRAIN_3,
  // end of track.
  PLAYER_STATE_EOT
};

struct _Player {
  Options options;
  volatile HWND hDlgConfig;
  volatile PlayerState state;

  struct {
    HINSTANCE hModule;
    char *pszModuleName;
    const char *pszFileName;
    double qVolume;
    HANDLE hEvent;

    struct {
      HANDLE hRead;
#if defined (YASAPI_CHECK_UNDERFLOW) // {
      HANDLE hUnderflow;
#endif // }
    } timer;

    PlayerStub *pStub;
  } base;

  struct {
    IMMDeviceEnumerator *pEnumerator;
#if defined (YASAPI_NOTIFY) // {
    IMMNotificationClient *pNotify;
#endif // }
  } run;

  PlayerDevice device;

  struct {
    const Strategy *pStrategy;
    WAVEFORMATEXTENSIBLE wfxx;
    WORD wBytesPerSample;
#if defined (YASAPI_SORROUND) // {
    int bSorround;
#endif // }
    int nShift;
    AUDCLNT_SHAREMODE eShareMode;
    DWORD dwStreamFlags;
    REFERENCE_TIME hnsDevicePeriod;
    Ring ring;
#if defined (YASAPI_RING_REALLOC) // {
    DWORD dwBufSize;
#endif // }
  } open;

  Connection connect;
#if defined (YASAPI_TIME_SEGMENT) // {
  Time time;
#else // } {
  Time offset1;
#if defined (YASAPI_OFFSET2) // {
  Time offset2;
#endif // }
  Time current;
#endif // }
};

#if 0 // {
int PlayerCreate(Player *pPlayer, HINSTANCE hModule, const wchar_t *path);
void PlayerDestroy(Player *pPlayer);
void PlayerKill(Player *pPlayer, int bReset, const PlayerState state);

int64_t PlayerRun(Player *pPlayer, Request *pRequest);
int64_t PlayerStop(Player *pPlayer, Request *pRequest);
int64_t PlayerOpen(Player *pPlayer, Request *pRequest);
#if defined (YASAPI_GAPLESS) // {
int64_t PlayerReset(Player *pPlayer, Request *pRequest);
#endif // }
int64_t PlayerClose(Player *pPlayer, Request *pRequest);
int64_t PlayerCanWrite(Player *pPlayer, Request *pRequest);
int64_t PlayerWrite(Player *pPlayer, Request *pRequest);
int64_t PlayerFlush(Player *pPlayer, Request *pRequest);
int64_t PlayerIsPlaying(Player *pPlayer, Request *pRequest);
int64_t PlayerPause(Player *pPlayer, Request *pRequest);
int64_t PlayerGetTime(Player *pPlayer, Request *pRequest);
int64_t PlayerRead(Player *pPlayer, Request *pRequest);
#if defined (YASAPI_CHECK_UNDERFLOW) // {
int64_t PlayerShutdownUnderflow(Player *pPlayer, Request *pRequest);
#endif // }
#if defined (YASAPI_NOTIFY) // {
int64_t PlayerMigrate(Player *pPlayer, Request *pRequest);
#endif // }

int64_t PlayerSend(Player *pPlayer, int exit, PlayerProc *pPlayerProc, ...);
void PlayerPost(Player *pPlayer, int exit, PlayerProc *pPlayerProc);
void PlayerPostRead(void *data);

UINT32 PlayerGetFramesPadding(Player *pPlayer);

HWND PlayerGetDlgConfig(Player *pPlayer);
void PlayerPostUpdate(Player *pPlayer, WORD wRing, WORD wShared);

void PlayerPostRead(void *data);
VOID CALLBACK PlayerFireRead(LPVOID lpArg, DWORD dwLow, DWORD dwHigh);
void PlayerSetTimerRead(Player *pPlayer, LONGLONG time);
#if defined (YASAPI_CHECK_UNDERFLOW) // {
void PlayerSetTimerCheckUnderflow(Player *pPlayer);
#endif // }

DWORD WINAPI PlayerThread(LPVOID lpParameter);

void PlayerCopyMemory(PVOID p, PVOID Destination, const VOID *Source,
    SIZE_T Length);

#if defined (YASAPI_NOTIFY) // {
int PlayerAddNotify(Player *pPlayer);
void PlayerRemoveNotify(Player *pPlayer);
#endif // }
#else // } {
int PlayerCreate(Player *pPlayer, HINSTANCE hModule, const wchar_t *path);
void PlayerDestroy(Player *pPlayer);
int64_t PlayerRun(Player *pPlayer, Request *pRequest);
int64_t PlayerStop(Player *pPlayer, Request *pRequest);
#if ! defined (YASAPI_TIME_SEGMENT) // {
int PlayerResetTime(Player *pPlayer, int bOffset);
#endif // }
int PlayerConnect(Player *pPlayer, int bNegociate, int bReset);
int PlayerCreateWFXX(Player *pPlayer, Request *pRequest);
int PlayerCreateRing(Player *pPlayer);
int PlayerReallocRing(Player *pPlayer);
int64_t PlayerGetMaxLatency(Player *pPlayer);
int64_t PlayerOpen(Player *pPlayer, Request *pRequest);
#if defined (YASAPI_NOTIFY) // {
int64_t PlayerMigrate(Player *pPlayer, Request *pRequest);
#endif // }
void PlayerShutdown(Player *pPlayer);
void PlayerDisconnect(Player *pPlayer);
void PlayerKill(Player *pPlayer, int bReset, const PlayerState state);
#if defined (YASAPI_GAPLESS) // {
int64_t PlayerReset(Player *pPlayer, Request *pRequest);
#endif // }
int64_t PlayerClose(Player *pPlayer, Request *pRequest);
int64_t PlayerCanWrite(Player *pPlayer, Request *pRequest);
int64_t PlayerPlay(Player *pPlayer, UINT32 uFramesPadding);
int PlayerTryStart(Player *pPlayer);
int64_t PlayerWrite(Player *pPlayer, Request *pRequest);
int64_t PlayerFlush(Player *pPlayer, Request *pRequest);
int64_t PlayerIsPlaying(Player *pPlayer, Request *pRequest);
int64_t PlayerPause(Player *pPlayer, Request *pRequest);
int64_t PlayerGetTime(Player *pPlayer, Request *pRequest);
#if defined (YASAPI_RING_REALLOC) // {
void PlayerWriteArgs(Player *pPlayer, Result *pResult, ...);
#endif // }
int64_t PlayerRead(Player *pPlayer, Request *pRequest);
#if defined (YASAPI_CHECK_UNDERFLOW) // {
int PlayerShutdownUnderflow(Player *pPlayer, Request *pRequest);
#endif // }
UINT32 PlayerGetFramesPadding(Player *pPlayer);
HWND PlayerGetDlgConfig(Player *pPlayer);
void PlayerPostUpdate(Player *pPlayer, WORD wRing, WORD wShared);
int64_t PlayerSend(Player *pPlayer, int exit, PlayerProc *pPlayerProc, ...);
void PlayerPost(Player *pPlayer, int exit, PlayerProc *pPlayerProc);
void PlayerPostRead(void *data);
VOID CALLBACK PlayerFireRead(LPVOID lpArg, DWORD dwLow, DWORD dwHigh);
void PlayerSetTimerRead(Player *pPlayer, LONGLONG time);
#if defined (YASAPI_CHECK_UNDERFLOW) // {
VOID CALLBACK PlayerFireShutdownUnderflow(LPVOID lpArg, DWORD dwLow,
    DWORD dwHigh);
void PlayerSetTimerCheckUnderflow(Player *pPlayer);
#endif // }
HANDLE PlayerGetEvent(Player *pPlayer);
DWORD WINAPI PlayerThread(LPVOID lpParameter);
void PlayerCopyMemory(PVOID p, PVOID Destination, const VOID *Source,
    SIZE_T Length);
#endif // }

///////////////////////////////////////////////////////////////////////////////
struct _Strategy {
  HRESULT (*Initialize)(Player *pPlayer);
  LONGLONG (*Interval)(Player *pPlayer, LONGLONG time);
  void (*SetTimer)(Player *pPlayer, LONGLONG time, BOOL bFlush);
  BOOL (*NeedDevicePeriod)(Player *pPlayer);
  BOOL (*NeedPadding)(Player *pPlayer);
  HRESULT (*SetEvent)(Player *pPlayer);
  HANDLE (*GetEvent)(Player *pPlayer);
  UINT32 (*GetFramesPadding)(Player *pPlayer);
};

extern const Strategy gcStrategyPush;
extern const Strategy gcStrategyPull;

///////////////////////////////////////////////////////////////////////////////
enum {
  WM_CONFIG_UPDATE=WM_APP
};

int ConfigDialog(Player *pPlayer, HINSTANCE hInstance, HWND hWndParent);
#if defined (YASAPI_ABOUT) // {
void AboutDialog(Player *pPlayer, HINSTANCE hInstance, HWND hWndParent);
#else // } {
void AboutDialog(HWND hWndParent);
#endif // }

#ifdef __cpluplus
}
#endif
#endif // }
